import logging
import os
import pwd
import re
import uuid
from dataclasses import dataclass, asdict
from enum import Enum

from clcommon.cpapi import cpinfo
from typing import List

from clwpos.feature_suites import get_allowed_modules, get_admin_suites_config, write_suites_allowed
from clwpos.optimization_features import OBJECT_CACHE_FEATURE, CDN_FEATURE, Feature


class BillingFeature(Enum):
    """
    Backwards-compatible list of features that we bill for.
    """
    ACCELERATE_WP_PREMIUM = 'AccelerateWP Premium'
    ACCELERATE_WP_CDN = 'AccelerateWP CDN Free'
    ACCELERATE_WP_CDN_PRO = 'AccelerateWP CDN Pro'


FEATURE_TO_BILLING_FEATURE = {
    OBJECT_CACHE_FEATURE: BillingFeature.ACCELERATE_WP_PREMIUM,
    CDN_FEATURE: BillingFeature.ACCELERATE_WP_CDN,
}

# constant returned by provision server if payment is ok
PAYMENT_OK_IDENTIFIER = 'ACTIVE'

def billing_feature_by_awp_feature(feature, allowed_suites):
    if feature != CDN_FEATURE:
        return FEATURE_TO_BILLING_FEATURE.get(feature)
    # because CND feature is included to multiple suites
    if CDNSuitePro.name in allowed_suites:
        return BillingFeature.ACCELERATE_WP_CDN_PRO
    return FEATURE_TO_BILLING_FEATURE.get(feature)

def is_valid_uuid(value):
    """
    Checks that string has uuid4 format
    """
    if value is None:
        return False
    return re.match('[0-9a-fA-F]{8}-[0-9a-fA-F]{4}-[0-9a-fA-F]{4}-[0-9a-fA-F]{4}-[0-9a-fA-F]{12}', value) is not None

def get_or_create_unique_identifier(username):
    """
    We need some unique identifier which user can
    use as his token on our provisioning server.

    We use uuid4 to make it somehow hard to bruteforce
    and still unique (hopefully as we don't check that :P)
    """
    if os.geteuid():
        return get_unique_identifier_as_user()

    pw = pwd.getpwnam(username)
    config = get_admin_suites_config(pw.pw_uid)

    unique_id = config.unique_id
    if unique_id is None or not is_valid_uuid(unique_id):
        config.unique_id = str(uuid.uuid4())
        write_suites_allowed(pw.pw_uid, pw.pw_gid, config)
    return config.unique_id


@dataclass
class UserRecord:
    username: str
    primary_domain: str
    id: str
    features: List[BillingFeature]


def get_report():
    """
    Collect report about billable users.

    Look for all users with allowed feature and add them
    to list for the further processing on CLN side.
    """
    report = []
    for user, domain in cpinfo(keyls=('cplogin', 'dns')):
        try:
            uid = pwd.getpwnam(user).pw_uid
        except KeyError:
            logging.warning('User %s does not have system uid; Malformed control panel configuration?')
            continue
        modules: List[Feature] = get_allowed_modules(uid)

        billable_features = []
        for module in modules:
            billing_feature = FEATURE_TO_BILLING_FEATURE.get(module)
            if billing_feature is None:
                continue
            billable_features.append(billing_feature.value)

        report.append(asdict(UserRecord(
            username=user,
            primary_domain=domain,
            id=get_or_create_unique_identifier(user),
            features=billable_features
        )))
    return report


def get_unique_identifier_as_user():
    """
    Get unique identifier for current end-user
    """
    from clwpos.utils import daemon_communicate
    from clwpos.daemon import WposDaemon

    unique_id = daemon_communicate({
        "command": WposDaemon.DAEMON_GET_UNIQUE_ID_COMMAND
    })["unique_id"]
    return unique_id
